﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using NEAT;

namespace RetinaProblem
{
    /// <summary>
    /// Main program for Retina Left and Right problem solved with HyperNEAT
    /// </summary>
    class Program
    {

        public static NEAT.HyperNEAT.Substrate sub;
        static double maxPct = 0.0;
        static System.IO.StreamWriter outputFile = null;
        public static int alternatingCount = 0;
        static bool AND = true;
        static double[] inputs = null;
        static int[] leftObjects = null;
        static int[] rightObjects = null;
        static double[,] possibleInputs = null;

        public static void Main(string[] args)
        {

            outputFile = new System.IO.StreamWriter("Output" + args[0] + ".txt");
               
            Run(args[0]);               

        }

        public static double thresh = 0.0;
        private static void Run(string paramName)
        {

            Evolution evo = new Evolution();
            evo.SetParameters(paramName + ".xml");
            evo.Evaluator = new NEAT.EvaluateGenomes(Program.RetinaEvaluatePopulation);
            thresh = NEAT.HyperNEAT.Substrate.param.StaticThreshold;
            evo.Initialise();
            
            evo.AllGenerations();

        }

        private static List<NEAT.Genetics.NodeGene> RetinaNodes()
        {

            List<NEAT.Genetics.NodeGene> Nodes = new List<NEAT.Genetics.NodeGene>();

            NEAT.HyperNEAT.SubstrateNode Node = new NEAT.HyperNEAT.SubstrateNode();
            int i = 1;

            Node.Function = "Identity";
            Node.Id = 0;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Bias;
            Node.SetCoordinate(-1, -1, -1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(-1, -1, 1);
            Nodes.Add(Node);


            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(-1, 1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(-0.33, -1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(-0.33, 1, 1);
            Nodes.Add(Node);


            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(0.33, -1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(0.33, 1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(1, -1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "Identity";
            Node.Id = i++;
            Node.Layer = 0;
            Node.Type = NEAT.Genetics.NodeType.Input;
            Node.SetCoordinate(1, 1, 1);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-1, -1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-1, 1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-0.33, -1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-0.33, 1, 0.75);
            Nodes.Add(Node);


            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(0.33, -1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(0.33, 1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(1, -1, 0.75);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 1;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(1, 1, 0.75);
            Nodes.Add(Node);


            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 2;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-1, 0, 0.5);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 2;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-0.33, 0, 0.5);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 2;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(0.33, 0, 0.5);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 2;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(1, 0, 0.5);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 3;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(-0.5, 0, 0.25);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 3;
            Node.Type = NEAT.Genetics.NodeType.Hidden;
            Node.SetCoordinate(0.5, 0, 0.25);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 4;
            Node.Type = NEAT.Genetics.NodeType.Output;
            Node.SetCoordinate(0, -1, 0);
            Nodes.Add(Node);

            Node = new NEAT.HyperNEAT.SubstrateNode();
            Node.Function = "BipolarSigmoid";
            Node.Id = i++;
            Node.Layer = 4;
            Node.Type = NEAT.Genetics.NodeType.Output;
            Node.SetCoordinate(0, 1, 0);
            Nodes.Add(Node);


            return Nodes;

        }

        /// <summary>
        /// Evaluates a population of genomes on the retina problems
        /// </summary>
        /// <param name="genomes">List of genomes</param>
        /// <param name="test">Indicates whether this is evolution, or testing</param>
        /// <returns></returns>
        private static bool RetinaEvaluatePopulation(List<NEAT.Genetics.NetworkGenome> genomes, bool test)
        {
            if (sub == null)
            {
                sub = new NEAT.HyperNEAT.Substrate();
                sub.GenerateNodes(RetinaNodes);
                sub.GenerateLinks(RetinaLinks);
            }
            bool success = false;
            maxPct = 0.0;
            for (int i = 0; i < genomes.Count; i++)
            {
                RetinaEvaluation(genomes[i], test);
                success = success || maxPct == 1;
            }

            outputFile.Write(maxPct + "\t");

            outputFile.Flush();

            alternatingCount++;
            if (alternatingCount == 500)
            {
                //AND = !AND;
                System.GC.Collect();
                System.GC.WaitForPendingFinalizers();
                alternatingCount = 0;
            }
            return success;

        }

        /// <summary>
        /// Generates links for the retina substrate
        /// </summary>
        /// <param name="sub">TGhe substrate links are being generated on</param>
        private static void RetinaLinks(NEAT.Genetics.NetworkGenome sub)
        {

            IEnumerable<NEAT.Genetics.NodeGene> incoming;

            int linkId = 0;
            for (int i = 0; i < sub.Nodes.Count; i++)
            {
                if (sub.Nodes[i].Type == NEAT.Genetics.NodeType.Input || sub.Nodes[i].Type == NEAT.Genetics.NodeType.Bias) continue;
                NEAT.HyperNEAT.SubstrateNode node = sub.Nodes[i] as NEAT.HyperNEAT.SubstrateNode;
                incoming = from n in sub.Nodes where n.Layer == node.Layer - 1 select n;
                foreach (NEAT.HyperNEAT.SubstrateNode n in incoming)
                {

                    sub.Links.Add(new NEAT.Genetics.LinkGene(linkId++, n.Id, node.Id, 0));

                }

            }
        }

        public static int crossConnections = 0;
        /// <summary>
        /// LEO methods of populating the substrate weight values
        /// </summary>
        /// <param name="sub">The substrate being populated</param>
        /// <param name="cppn">The CPPN generating the weight values</param>
        private static void RetinaCreate(NEAT.Genetics.NetworkGenome sub, NEAT.Networks.DecodedNetwork cppn)
        {

            IEnumerable<NEAT.Genetics.NodeGene> incoming;
            double[] inputs = new double[6];
            double[] outputs;
            int linkId = 0;
            for (int i = 0; i < sub.Links.Count; i++)
            {
                NEAT.HyperNEAT.SubstrateNode target = sub.Nodes[sub.Links[i].Target] as NEAT.HyperNEAT.SubstrateNode;
                NEAT.HyperNEAT.SubstrateNode source = sub.Nodes[sub.Links[i].Source] as NEAT.HyperNEAT.SubstrateNode;
                if (sub.Links[i].Source == 0)
                {
                    cppn.Flush();
                    inputs[0] = target.Coordinate[0];
                    inputs[1] = target.Coordinate[1];
                    inputs[2] = target.Coordinate[2];
                    inputs[3] = target.Coordinate[0];
                    inputs[4] = target.Coordinate[1];
                    inputs[5] = target.Coordinate[2];
                    cppn.SetInputs(inputs);

                    cppn.ActivateNetwork(5);

                    outputs = cppn.GetOutputs();
                    sub.Links[i].Weight = outputs[1];

                }
                else
                {

                    cppn.Flush();
                    inputs[0] = source.Coordinate[0];
                    inputs[1] = source.Coordinate[1];
                    inputs[2] = source.Coordinate[2];
                    inputs[3] = target.Coordinate[0];
                    inputs[4] = target.Coordinate[1];
                    inputs[5] = target.Coordinate[2];
                    cppn.SetInputs(inputs);

                    cppn.ActivateNetwork(5);

                    outputs = cppn.GetOutputs();
                    sub.Links[i].Weight = outputs[0];
                                       
                    if (outputs[2] <  thresh ) { sub.Links[i].Weight = 0; }
                    

                }

            }

        }

        /// <summary>
        /// Evaluates a single genome on the retina problem
        /// </summary>
        /// <param name="cppn">CPPN generating the substrate weights</param>
        /// <param name="test">Indicates evolution or testing</param>
        private static void RetinaEvaluation(NEAT.Genetics.NetworkGenome cppn, bool test)
        {
            crossConnections = 0;
            #region Evaluation
            if (!test)
            {
                NEAT.Networks.DecodedNetwork net = sub.GenerateNetwork(RetinaCreate, NEAT.Networks.DecodedNetwork.DecodeGenome(cppn));

                #region Set Input Arrays
                if (possibleInputs == null)
                {
                    inputs = new double[8];

                    leftObjects = new int[8];
                    rightObjects = new int[8];
                    int lefts = 0;
                    int rights = 0;
                    possibleInputs = new double[16, 4];
                    int p = 0;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = -3;
                    leftObjects[lefts++] = p;
                    rightObjects[rights++] = p;
                    p++;


                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = -3;
                    leftObjects[lefts++] = p;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = -3;
                    leftObjects[lefts++] = p;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = -3;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = -3;
                    leftObjects[lefts++] = p;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = -3;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = -3;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = -3;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = 3;
                    leftObjects[lefts++] = p;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = 3;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = 3;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = -3;
                    possibleInputs[p, 3] = 3;
                    rightObjects[rights++] = p;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = 3;
                    leftObjects[lefts++] = p;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = -3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = 3;
                    leftObjects[lefts++] = p;
                    p++;

                    possibleInputs[p, 0] = -3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = 3;
                    leftObjects[lefts++] = p;
                    p++;

                    possibleInputs[p, 0] = 3;
                    possibleInputs[p, 1] = 3;
                    possibleInputs[p, 2] = 3;
                    possibleInputs[p, 3] = 3;

                }

                #endregion

                bool isLeftObject = false;
                bool isRightObject = false;

                double[] outputs;

                double correct = 0.0;
                cppn.Fitness = 0;
                double fitness = 0.0;

                for (int i = 0; i < 16; i++)
                {
                    for (int j = 0; j < 16; j++)
                    {


                        isLeftObject = leftObjects.Contains(i);
                        isRightObject = rightObjects.Contains(j);

                        for (int k = 0; k < 4; k++)
                        {
                            inputs[k] = possibleInputs[i, k];

                        }

                        for (int k = 0; k < 4; k++)
                        {
                            inputs[k + 4] = possibleInputs[j, k];
                        }

                        net.SetInputs(inputs);

                        net.ActivateNetwork(5);

                        outputs = net.GetOutputs();
                        outputs[0] = Math.Max(outputs[0], -1.0);
                        outputs[0] = Math.Min(outputs[0], 1.0);
                        outputs[1] = Math.Max(outputs[1], -1.0);
                        outputs[1] = Math.Min(outputs[1], 1.0);

                        if (isLeftObject && isRightObject)
                        {
                            if (outputs[0] > 0 && outputs[1] > 0)
                            {
                                correct += 1;
                            }

                            fitness += Math.Abs(1.0 - outputs[0]);
                            fitness += Math.Abs(1.0 - outputs[1]);
                        }
                        else if (isLeftObject)
                        {
                            if (outputs[0] > 0 && outputs[1] <= 0)
                            {
                                correct += 1;
                            }

                            fitness += Math.Abs(1.0 - outputs[0]);

                            fitness += Math.Abs(-1.0 - outputs[1]);
                        }
                        else if (isRightObject)
                        {


                            if (outputs[0] <= 0 && outputs[1] > 0)
                            {
                                correct += 1;
                            }

                            fitness += Math.Abs(-1.0 - outputs[0]);

                            fitness += Math.Abs(1.0 - outputs[1]);

                        }
                        else
                        {
                            if (outputs[0] <= 0 && outputs[1] <= 0)
                            {
                                correct += 1;
                            }

                            fitness += Math.Abs(-1.0 - outputs[0]);

                            fitness += Math.Abs(-1.0 - outputs[1]);
                        }

                        net.Flush();

                    }

                }

                maxPct = Math.Max(maxPct, correct / 256.0);

                cppn.Fitness = 1000.0 / (1.0 + fitness * fitness);
            }
            #endregion
            else
            {
               NEAT.Networks.DecodedNetwork net = sub.GenerateNetwork(RetinaCreate, NEAT.Networks.DecodedNetwork.DecodeGenome(cppn));

                outputFile.WriteLine(crossConnections);

            }

        }

    }
}
